#include "stdafx.h"

#include "ModelReport.h"

const double g_x=M_PI;

//TODO - MODE_ should be an enum 
const int MODE_VIEW = 0;
const int MODE_NEWSTATE = 1;
const int MODE_NEWRATE1 = 2;
const int MODE_NEWRATE2 = 3;
const int MODE_NEWCONSTRAINT = 4;
const int MODE_EDITSTATE = 5;
const int MODE_EDITRATE = 6;
const int MODE_EDITCONSTRAINT = 7;
const int MODE_EDITCLASS = 8;

//-- Capture mode 
const int CAP_NONE = 0;
const int CAP_STATE = 1;
const int CAP_RATE = 2;
const int CAP_CONSTRAINT = 3;
const int CAP_CONSTRUCTRATE = 4;

const COLORREF CR_CONSTRAINTBGR = RGB( 128, 128, 255 );	
const COLORREF CR_CONSTRAINTTEXT = RGB( 0, 0, 0 );
const COLORREF CR_STATEDARK = RGB(0,0,0);
const COLORREF CR_STATELIGHT = RGB(255,255,255);
	
const int STATERADIUS = 10;

const int CLSEL_PERROW = 20;			// parameters for drawing class selection boxes 
const int CLSEL_DELTAX = 22;			// boxes per row, x/y offset between,
const int CLSEL_DELTAY = 22;
const int CLSEL_WIDTH = 19;				// size of box draw
const int CLSEL_HEIGHT = 19;		
const int CLSEL_EXT = 2;				// extra size for selection class outline
const int CLSEL_YOFFSET = 3 * CLSEL_DELTAY ;
const int CLSEL_XOFFSET = 2 ;			// location from lower left of properties dialogue to start first box

const int CONSTRAINTWIDTH = 60;
const int CONSTRAINTHEIGHT = 8;

const char * g_pchModelFilter = "Model file (*.qmf)\0*.qmf\0All Files (*.*)\0*.*\0\0";
const char * g_pchModelExtension = "qmf";

class CVwModel : public CDocumentDialog { 
	private :
		CModel m_oModel;
		//bool m_bNewFile;				// True on new file until saved (currently unused)
		
		COLORREF m_acrTraceColor[16];	// Color select - local color array 
		HBRUSH hbrButton;				// for coloring the selected class button

		int m_iNewSel;					// which item type will be appended if canvas is clicked
		int m_iMode;					// Which MODE_* is dialogue currently in 
		bool m_bUndoAvailable;			// changes have been made since the last save operation
		int m_iRateStartState;			// During rate building, the first selected state

		//-- 
		RECT m_rctProperties;			// area of properties panel 
		int m_iClLeft, m_iClTop;		// Location of class selection boxes during class edit.
		
		//-- canvas for drawing model 
		RECT m_rctCanvas;				// coordinates of display area for model 
		int m_iXSize,m_iYSize;			// pre-calc'd for efficiency - width, height, midpoint of canvas.
		int m_iXMid,m_iYMid;			
		int XCanvas( double dX ) {		// convert +/-100.0 coordinates to canvas coordinates
			return m_iXMid + int( dX / 200.0f * m_iXSize );
			}
		int YCanvas( double dY ) {		// convert +/-100.0 coordinates to canvas coordinates
			return m_iYMid + int( dY / 200.0f * m_iYSize );
			}
		double XPosition( int iX ) { 
			return max( -100.0, min(100.0, double(MulDiv(iX-m_rctCanvas.left,200,m_rctCanvas.right-m_rctCanvas.left))-100.0 ) ) ;
			}
		double YPosition( int iY ) { 
			return max( -100.0, min(100.0, double(MulDiv(iY-m_rctCanvas.top,200,m_rctCanvas.bottom-m_rctCanvas.top))-100.0 ) ) ;
			}

		int m_iRateTextHeight;			// size of rate text for validating clicks on rates
		
		//-- CapType and CapItem indicate the 'current' item.  When the current item is 
		// being dragged, m_pCapture!=NULL and m_iCapX, m_iCapY are the location when 
		// the drag operation started.  m_dCapX and m_dCapY are the original location of
		// the object being dragged.
		double m_dCapX, m_dCapY;		// Location of state or constraint when 'captured'
		int m_iCapX, m_iCapY;			// Location of mouse click when 'captured'
		int m_iCapType;					// 0=none 1=state 2=rate 3=constraint
		int m_iCapItem;					// Captured item index in array 
		int m_iSelClass;				// during class edit - which class is being modified
		
		CDlgCheck m_selClass;
		CDlgCheck m_selRate;
		CDlgCheck m_selConstraint;

		CDlgControl m_lblRateK0;
		CDlgControl m_lblRateMult1;
		CDlgControl m_lblRateP;
		CDlgControl m_lblRateExp;
		CDlgControl m_lblRateLParen;
		CDlgControl m_lblRateK1;
		CDlgControl m_lblRateMult2;
		CDlgControl m_lblRateQ;
		CDlgControl m_lblRateRParen;
		CDlgControl m_lblRateKFwd;
		CDlgControl m_lblRateFwd;
		CDlgEdit m_edtRateK0F;
		CDlgCheck m_edtRatePF;
		CDlgEdit m_edtRatePNameF;
		CDlgEdit m_edtRateK1F;
		CDlgCheck m_edtRateQF;
		CDlgEdit m_edtRateQNameF;
		CDlgControl m_lblRateKBwd;
		CDlgControl m_lblRateBwd;
		CDlgEdit m_edtRateK0B;
		CDlgCheck m_edtRatePB;
		CDlgEdit m_edtRatePNameB;
		CDlgEdit m_edtRateK1B;
		CDlgCheck m_edtRateQB;
		CDlgEdit m_edtRateQNameB;
		CDlgControl m_lblRatePName;
		CDlgControl m_lblRateQName;

		CDlgControl m_lblClassColor;
		CDlgButton m_btnClassColor;
		CDlgControl m_lblClassAmp;
		CDlgEdit m_edtClassAmp;
		CDlgControl m_lblClassStd;
		CDlgEdit m_edtClassStd;
		CDlgControl m_lblClassCond;
		CDlgEdit m_edtClassCond;
		CDlgButton m_btnClassAdd;

		CDlgControl m_lblState;
		CDlgControl m_lblStateClass;
		CDlgCombo m_cmbStateClass;
		CDlgControl m_lblStateStartProb;
		CDlgEdit m_edtStateStartProb;
		CDlgControl m_lblStateAmp;
		CDlgEdit m_edtStateAmp;
		CDlgControl m_lblStateStd;
		CDlgEdit m_edtStateStd;

		CDlgControl m_lblCnsType;
		CDlgControl m_lblCnsRate;
		CDlgControl m_lblCnsRefRate;
		CDlgCombo m_cmbCnsType;
		CDlgCombo m_cmbCnsRate;
		CDlgCombo m_cmbCnsRefRate;

		CDlgButton m_btnClasses;
		CDlgButton m_btnSave;
		CDlgButton m_btnClose;
		CDlgButton m_btnDelete;
		CDlgButton m_btnEnergy;
		CDlgButton m_btnUndo;
		CDlgControl m_lblHelp;
		CDlgControl m_picCanvas;
		CDlgControl m_grpProperties;
		
	public : 
		//-- constructor
		CVwModel() : CDocumentDialog( g_pchModelFilter, g_pchModelExtension ) { 
			m_iID = IDD_MODELVIEW;
			m_pCapture=NULL;	// control when mouse is captured 
			m_iRateTextHeight = 1;
			m_iNewSel=-1;
			m_iMode=MODE_VIEW;
			m_bUndoAvailable=false;
			SetFileName("");
			hbrButton = NULL;
			//m_bNewFile=false;
			}
			
		//-- destructor
		~CVwModel() { 
			if( hbrButton != NULL ) 
				DeleteObject(hbrButton);
			}

		void SetFileName( const char * pchName ) { 
			m_cFileName.Set(pchName);
			SetWindowText( GetParent(m_hDlg), pchName );
			}

		void * GetData( int iData, void * pInfo ) { 
			if( iData == 1 ) 
				return m_cFileName.Buffer();
			if( iData == 2 ) 
				return &m_oModel;
			return NULL;
			}

		//=========================================================== Fill in controls
		//-- Enable or disable PName or QName dependent on checkbox status
		void EnableRateNames() { 
			m_edtRatePNameF.SetEnabled( m_edtRatePF.GetCheck() );
			m_edtRateQNameF.SetEnabled( m_edtRateQF.GetCheck() );
			m_edtRatePNameB.SetEnabled( m_edtRatePB.GetCheck() );
			m_edtRateQNameB.SetEnabled( m_edtRateQB.GetCheck() );
			}

		//-- 			
		void InitializeControls( ) { 
			//-- create initial color array for extra colors
			for( int i=0; i<16; ++i ) 
				m_acrTraceColor[i]=RGB(0,0,16*i);

			RECT rc;

			//-- get a few controls setup and find locations to use for moving controls.
			// For ease of layout, the three groups of controls that appear at the top of 
			// the dialogue are laid out non-overlapping in the editor then moved into 
			// position in this method.  This establishes how much to move each group of 
			// controls.
			m_lblRateK0.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEK0) );
			m_lblRateK0.GetControlBounds(&rc);
			int iRatePos = rc.top;
			
			m_lblClassColor.SetHandle( GetDlgItem( m_hDlg, IDC_LBLCLASSCOLOR) );
			m_lblClassColor.GetControlBounds(&rc);
			int iClassOffset = iRatePos-rc.top;

			m_lblState.SetHandle( GetDlgItem( m_hDlg, IDC_LBLSTATE) );
			m_lblState.GetControlBounds(&rc);
			int iStateOffset = iRatePos-rc.top;

			//--New item mode
			m_selClass.SetHandle( GetDlgItem( m_hDlg, IDC_SEL_CLASS), "Select to add a new state");
			m_selRate.SetHandle( GetDlgItem( m_hDlg, IDC_SEL_RATE), "Select to add a new rate");
			m_selConstraint.SetHandle( GetDlgItem( m_hDlg, IDC_SEL_CONSTRAINT), "Select to add a new constraint");

			//--Rate area headings
			m_lblRateMult1.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEMULT1) );
			m_lblRateP.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEP) );
			m_lblRateExp.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEEXP) );
			m_lblRateLParen.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATELPAREN) );
			m_lblRateK1.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEK1) );
			m_lblRateMult2.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEMULT2) );
			m_lblRateQ.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEQ) );
			m_lblRateRParen.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATERPAREN) );

			//--Rate area left titles
			m_lblRateKFwd.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEKFWD) );
			m_lblRateFwd.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEFWD) );
			m_lblRateKBwd.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEKBWD) );
			m_lblRateBwd.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEBWD) );

			//--Rate area edits (forward)
			m_edtRateK0F.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEK0F), "Pre-multipler (k0) for the selected rate");
			m_edtRatePF.SetHandle( GetDlgItem( m_hDlg, IDC_CHKRATEPF), "Toggle inclusion of an experimental condition in the base rate.");
			m_edtRatePNameF.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEPNAMEF), "Ligand name for the selected rate or blank if none.  Name should correspond to a value in the experimental conditions.");
			m_edtRateK1F.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEK1F), "Rate exponential component (k1) for the selected rate.  Generally used for voltage or pressure");
			m_edtRateQF.SetHandle( GetDlgItem( m_hDlg, IDC_CHKRATEQF), "Toggle inclusion of an experimental condition in the exponential portion of the rate.");
			m_edtRateQNameF.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEQNAMEF), "Exponential multipler name for the selected rate, or blank if none.  Name should correspond to a value in the experimental conditions.");

			//--Rate area edits (backward)
			m_edtRateK0B.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEK0B), "Pre-multipler (k0) for the reverse of the selected rate");
			m_edtRatePB.SetHandle( GetDlgItem( m_hDlg, IDC_CHKRATEPB), "Toggle inclusion of an experimental condition in the base rate.");
			m_edtRatePNameB.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEPNAMEB), "Ligand name for the reverse of the selected rate or blank if none.  Name should correspond to a value in the experimental conditions.");
			m_edtRateK1B.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEK1B), "Rate exponential component (k1) for the reverse of the selected rate.  Generally used for voltage or pressure");
			m_edtRateQB.SetHandle( GetDlgItem( m_hDlg, IDC_CHKRATEQB), "Toggle inclusion of an experimental condition in the exponential portion of the rate.");
			m_edtRateQNameB.SetHandle( GetDlgItem( m_hDlg, IDC_EDTRATEQNAMEB), "Exponential multipler name for the reverse of the selected rate, or blank if none.  Name should correspond to a value in the experimental conditions.");

			//--Rate area lower notes
			m_lblRatePName.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEPNAME) );
			m_lblRateQName.SetHandle( GetDlgItem( m_hDlg, IDC_LBLRATEQNAME) );
			EnableRateNames();

			//-- class area 
			m_btnClassColor.SetHandle( GetDlgItem( m_hDlg, IDC_BTNCLASSCOLOR), "Select the color to use for all states of this class.");
			m_lblClassAmp.SetHandle( GetDlgItem( m_hDlg, IDC_LBLCLASSAMP) );
			m_edtClassAmp.SetHandle( GetDlgItem( m_hDlg, IDC_EDTCLASSAMP), "Current for states of this class.");
			m_lblClassStd.SetHandle( GetDlgItem( m_hDlg, IDC_LBLCLASSSTD) );
			m_edtClassStd.SetHandle( GetDlgItem( m_hDlg, IDC_EDTCLASSSTD), "Standard deviation for this class.   For other than class 0, the class 0 deviation is added to this value.");
			m_lblClassCond.SetHandle( GetDlgItem( m_hDlg, IDC_LBLCLASSCOND) );
			m_edtClassCond.SetHandle( GetDlgItem( m_hDlg, IDC_EDTCLASSCOND), "Conditioning value for this class.");
			m_btnClassAdd.SetHandle( GetDlgItem( m_hDlg, IDC_BTN_ADDCLASS), "Add a new class.");

			m_lblClassColor.MoveWindow( 0, iClassOffset );
			m_btnClassColor.MoveWindow( 0, iClassOffset );
			m_lblClassAmp.MoveWindow( 0, iClassOffset );
			m_edtClassAmp.MoveWindow( 0, iClassOffset );
			m_lblClassStd.MoveWindow( 0, iClassOffset );
			m_edtClassStd.MoveWindow( 0, iClassOffset );
			m_lblClassCond.MoveWindow( 0, iClassOffset );
			m_edtClassCond.MoveWindow( 0, iClassOffset );
			
			//-- State area 
			m_lblStateClass.SetHandle( GetDlgItem( m_hDlg, IDC_LBLSTATECLASS) );
			m_cmbStateClass.SetHandle( GetDlgItem( m_hDlg, IDC_CMBSTATECLASS), "The class corresponding to the selected state.");
			m_lblStateStartProb.SetHandle( GetDlgItem( m_hDlg, IDC_LBLSTATESTARTPROB) );
			m_edtStateStartProb.SetHandle( GetDlgItem( m_hDlg, IDC_EDTSTATESTARTPROB), "(P0) The initial probability for the selected state. ");
			m_lblStateAmp.SetHandle( GetDlgItem( m_hDlg, IDC_LBLSTATEAMP) );
			m_edtStateAmp.SetHandle( GetDlgItem( m_hDlg, IDC_EDTSTATEAMP), "Current amplitude from the selected class properties.");
			m_lblStateStd.SetHandle( GetDlgItem( m_hDlg, IDC_LBLSTATESTD) );
			m_edtStateStd.SetHandle( GetDlgItem( m_hDlg, IDC_EDTSTATESTD), "Current standard deviation from the selected class properties.");
			
			m_cmbStateClass.ClearContents();
			m_cmbStateClass.AddString( "1" );
			m_cmbStateClass.AddString( "2" );
			m_cmbStateClass.AddString( "3" );
			m_cmbStateClass.AddString( "4" );
			m_lblState.MoveWindow( 0, iStateOffset );
			m_lblStateClass.MoveWindow( 0, iStateOffset );
			m_cmbStateClass.MoveWindow( 0, iStateOffset );
			m_lblStateStartProb.MoveWindow( 0, iStateOffset );
			m_edtStateStartProb.MoveWindow( 0, iStateOffset );
			m_lblStateAmp.MoveWindow( 0, iStateOffset );
			m_edtStateAmp.MoveWindow( 0, iStateOffset );
			m_lblStateStd.MoveWindow( 0, iStateOffset );
			m_edtStateStd.MoveWindow( 0, iStateOffset );
					
			//----- Constraint area 
			m_lblCnsType.SetHandle( GetDlgItem( m_hDlg, IDC_LBL_CNS_TYPE) );
			m_lblCnsRate.SetHandle( GetDlgItem( m_hDlg, IDC_LBL_CNS_RATE) );
			m_lblCnsRefRate.SetHandle( GetDlgItem( m_hDlg, IDC_LBL_CNS_REFRATE) );
			m_cmbCnsType.SetHandle( GetDlgItem( m_hDlg, IDC_CMB_CNS_TYPE) );
			m_cmbCnsRate.SetHandle( GetDlgItem( m_hDlg, IDC_CMB_CNS_RATE) );
			m_cmbCnsRefRate.SetHandle( GetDlgItem( m_hDlg, IDC_CMB_CNS_REFRATE) );
			//-- fill in Type combo
			m_cmbCnsType.ClearContents();
			m_cmbCnsType.AddString( sCONS_FIXRATE );
			m_cmbCnsType.AddString( sCONS_FIXEXP );
			m_cmbCnsType.AddString( sCONS_SCALERATE );
			m_cmbCnsType.AddString( sCONS_SCALEEXP );
			//-- Offset controls into correct area ( easier to edit dialogue in wrong spot. )
			m_lblCnsType.GetControlBounds(&rc);
			int iConstraintOffset = iRatePos-rc.top;
			m_lblCnsType.MoveWindow( 0, iConstraintOffset );
			m_lblCnsRate.MoveWindow( 0, iConstraintOffset );
			m_lblCnsRefRate.MoveWindow( 0, iConstraintOffset );
			m_cmbCnsType.MoveWindow( 0, iConstraintOffset );
			m_cmbCnsRate.MoveWindow( 0, iConstraintOffset );
			m_cmbCnsRefRate.MoveWindow( 0, iConstraintOffset );
		
			//-- buttons 
			m_btnSave.SetHandle( GetDlgItem( m_hDlg, IDC_BTNSAVE), "Save changes to this model and close the editor.");
			m_btnClose.SetHandle( GetDlgItem( m_hDlg, IDC_BTNCLOSE), "Close the model editor.");
			m_btnDelete.SetHandle( GetDlgItem( m_hDlg, IDC_BTNDELETE), "Delete the currently selected state, rate, or constraint.");
			m_btnEnergy.SetHandle( GetDlgItem( m_hDlg, IDC_BTNENERGY), "Open Energy landscape view.");
			m_btnUndo.SetHandle( GetDlgItem( m_hDlg, IDC_BTNUNDO), "Undo last change to the model.");
			m_btnClasses.SetHandle( GetDlgItem( m_hDlg, IDC_BTN_CLASSES), "Modify the conductance and variance for classes.");
			
			//--Other			
			m_lblHelp.SetHandle( GetDlgItem( m_hDlg, IDC_LBLHELP) );
			m_picCanvas.SetHandle( GetDlgItem( m_hDlg, IDC_CANVAS) );
			m_grpProperties.SetHandle( GetDlgItem( m_hDlg, IDC_GRPPROPERTIES ) ) ;
			m_grpProperties.GetControlBounds( &m_rctProperties );
			m_iClTop = m_rctProperties.bottom - CLSEL_YOFFSET;
			m_iClLeft = m_rctProperties.left + CLSEL_XOFFSET;
			SetWindowPos( m_btnClassAdd.GetHandle(), 0, m_rctProperties.right-24, m_rctProperties.bottom-24, 0, 0, SWP_NOREPOSITION|SWP_NOSIZE|SWP_NOACTIVATE );
			
			SetMode(MODE_VIEW);
			}

		//-- fill in list of classes based on model contents
		void SetClassList() { 
			char buf[16];
			m_cmbStateClass.ClearContents();
			for( int iClass=0; iClass<m_oModel.m_aClasses.Count(); ++iClass ) { 
				sprintf( buf, "%d", iClass );
				m_cmbStateClass.AddString(buf);
				}
			}
			
		//-- Set each controls visible / enable as appropriate 
		void SetMode( int iMode ) {
			//-- Edit Class mode does specialized drawing which requires an update
			//if( iMode==MODE_EDITCLASS || m_iMode==MODE_EDITCLASS ) // coming or going.
			// Above if() statement : m_iMode is sometimes set outside this func so prior 
			// mode might be deceptive.instead of checking, always invalidate.
			PropertiesInvalidate();
		
			//----- If changing out of an edit mode, save any changes
			if( m_iMode == MODE_EDITSTATE ) { 
				//TODO - save any changes 
				}
			if( m_iMode == MODE_EDITRATE ) { 
				//TODO - save any changes 
				}
			if( m_iMode == MODE_EDITCONSTRAINT ) { 
				//TODO - save any changes 
				}
				
			//----- store the new mode 
			m_iMode=iMode;

			//----- change controls visible / enabled status as appropriate
			//m_selClass.SetCheck( m_iMode == MODE_NEWSTATE );
			//m_selRate.SetCheck( m_iMode == MODE_NEWRATE1 );
			//m_selConstraint.SetCheck( m_iMode == MODE_NEWCONSTRAINT );

			m_lblRateK0.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateMult1.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateP.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateExp.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateLParen.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateK1.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateMult2.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateQ.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateRParen.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateKFwd.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateFwd.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateK0F.SetVisible( iMode==MODE_EDITRATE );
			m_edtRatePF.SetVisible( iMode==MODE_EDITRATE );
			m_edtRatePNameF.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateK1F.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateQF.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateQNameF.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateKBwd.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateBwd.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateK0B.SetVisible( iMode==MODE_EDITRATE );
			m_edtRatePB.SetVisible( iMode==MODE_EDITRATE );
			m_edtRatePNameB.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateK1B.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateQB.SetVisible( iMode==MODE_EDITRATE );
			m_edtRateQNameB.SetVisible( iMode==MODE_EDITRATE );
			m_lblRatePName.SetVisible( iMode==MODE_EDITRATE );
			m_lblRateQName.SetVisible( iMode==MODE_EDITRATE );

			m_lblClassColor.SetVisible( iMode==MODE_EDITCLASS );
			m_btnClassColor.SetVisible( iMode==MODE_EDITCLASS );
			m_lblClassAmp.SetVisible( iMode==MODE_EDITCLASS );
			m_edtClassAmp.SetVisible( iMode==MODE_EDITCLASS );
			m_lblClassStd.SetVisible( iMode==MODE_EDITCLASS );
			m_edtClassStd.SetVisible( iMode==MODE_EDITCLASS );
			m_lblClassCond.SetVisible( iMode==MODE_EDITCLASS );
			m_edtClassCond.SetVisible( iMode==MODE_EDITCLASS );
			m_btnClassAdd.SetVisible( iMode==MODE_EDITCLASS && m_oModel.m_aClasses.Count()<MAXCLASSES );

			m_lblState.SetVisible( iMode==MODE_EDITSTATE );
			m_lblStateClass.SetVisible( iMode==MODE_EDITSTATE );
			m_cmbStateClass.SetVisible( iMode==MODE_EDITSTATE );
			m_lblStateStartProb.SetVisible( iMode==MODE_EDITSTATE );
			m_edtStateStartProb.SetVisible( iMode==MODE_EDITSTATE );
			m_lblStateAmp.SetVisible( iMode==MODE_EDITSTATE );
			m_edtStateAmp.SetVisible( iMode==MODE_EDITSTATE );
			m_lblStateStd.SetVisible( iMode==MODE_EDITSTATE );
			m_edtStateStd.SetVisible( iMode==MODE_EDITSTATE );

			m_lblCnsType.SetVisible( iMode==MODE_EDITCONSTRAINT );
			m_lblCnsRate.SetVisible( iMode==MODE_EDITCONSTRAINT );
			m_lblCnsRefRate.SetVisible( iMode==MODE_EDITCONSTRAINT );
			m_cmbCnsType.SetVisible( iMode==MODE_EDITCONSTRAINT );
			m_cmbCnsRate.SetVisible( iMode==MODE_EDITCONSTRAINT );
			m_cmbCnsRefRate.SetVisible( iMode==MODE_EDITCONSTRAINT );
			
			m_btnSave.SetEnabled( true );	
			m_btnClose.SetEnabled( true );	
			m_btnClasses.SetEnabled( true );
			m_btnDelete.SetEnabled( iMode==MODE_EDITRATE || iMode==MODE_EDITCLASS || iMode==MODE_EDITCONSTRAINT || iMode==MODE_EDITSTATE );
			m_btnEnergy.SetEnabled( true );
			m_btnUndo.SetEnabled( m_bUndoAvailable );	

			//-- Changes specific to the new mode, such as mode dependant help text 
			m_lblHelp.SetVisible( true );
			switch( iMode ) { 
				case MODE_VIEW :
					m_lblHelp.SetText("Click on a rate, state, or constraint to alter it.  Click and drag a rate or constraint to move it.");
					break;
				case MODE_NEWSTATE :
					m_lblHelp.SetText("Click in the model display panel below to add a new state.");
					break;
				case MODE_NEWRATE1 :
					m_lblHelp.SetText("Add a new rate : Click on the initial state and drag the mouse to the final state.");
					break;
				case MODE_NEWRATE2 :
					m_lblHelp.SetText("Drag the mouse to the final state to complete the new rate.");
					break;
				case MODE_NEWCONSTRAINT :
					m_lblHelp.SetText("Click in the model display panel below to add a new constraint.");
					break;
				case MODE_EDITSTATE : 
					m_lblHelp.SetText("Change the values above to modify the selected state.  Click on [Delete] to remove it.");
					break;
				case MODE_EDITRATE :
					m_lblHelp.SetText("Change the values above to modify the selected rate.  Click on [Delete] to remove it.");
					break;
				case MODE_EDITCONSTRAINT :
					m_lblHelp.SetText("Change the values above to modify the selected constraint.  Click on [Delete] to remove it.");
					break;
				case MODE_EDITCLASS	:
					m_lblHelp.SetText("Click on a class box to select the class to edit. Change the values above to modify the class properties.");
					break;
				default : 
					m_lblHelp.SetVisible( false );
					return;
				}
			}

		//=========================================================== Drawing
		
		//-- Create text for rate   Ex :  12.34 pq 1.234
		void RateText( char * pch, double dK0, double dK1, bool bPDep, bool bQDep ) { 
			//-- K0
			if( dK0 > 100000000.0 || dK0 < .0001 ) 
				sprintf( pch,  "%g", dK0 );
			else if(  dK0 > 1000.0 )
				sprintf( pch,  "%.0f", dK0 );
			else if(  dK0 > 10.0 )
				sprintf( pch,  "%.2f", dK0 );
			else if(  dK0 > 0.1 )
				sprintf( pch,  "%.4f", dK0 );
			else 
				sprintf( pch,  "%.6f", dK0 );
				
			// P and Q 
			if( ! bQDep ) {
				if( bPDep )
					strcat( pch, " p" );			
				}
			else {
				strcat( pch, bPDep ? " pq " : " q " );			
			
				// k1 
				if( dK1 > 100000000.0 || dK1 < .0001 ) 
					sprintf( pch+strlen(pch),  "%g", dK1 );
				else if(  dK1 > 1000.0 )
					sprintf( pch+strlen(pch),  "%.0f", dK1 );
				else if(  dK1 > 10.0 )
					sprintf( pch+strlen(pch),  "%.2f", dK1 );
				else if(  dK1 > 0.1 )
					sprintf( pch+strlen(pch),  "%.4f", dK1 );
				else 
					sprintf( pch+strlen(pch),  "%.6f", dK1 );
				}
				
			}
		
		//-- cause redraw of the model building pane
		void CanvasInvalidate() { 
			InvalidateRect( m_hDlg, &m_rctCanvas, true );
			}

		//-- Cause redraw of the properties edit pane
		void PropertiesInvalidate() { 
			InvalidateRect( m_hDlg, NULL, true );
			}

		//--
		void OnPaint(  ) { 
			//--- Don't continuously redraw while sizing
			//if( m_bSizing ) 
			//	return;
								
			//------------------------------------------------------ Update the display
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(m_hDlg, &ps);

			IntersectClipRect( hdc, m_rctCanvas.left, m_rctCanvas.top, m_rctCanvas.right, m_rctCanvas.bottom );
			
			//--  text drawing prep.						
			int iBkModeWas = SetBkMode( hdc, TRANSPARENT );
			char bufText[256];			// buffer for text output
			HFONT fntState = CreateFont( //TODO - doc suggests for MM_TEXT, nHeight = -MulDiv(PointSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);
									-12 /*Height*/,		0 /*Width*/, 
									0/*escapement*/,	0/*Orientation*/,  
									0/*Weight*/,		FALSE/*Italic*/,
									FALSE/*Underline*/,	FALSE/*Strikeout*/,
									ANSI_CHARSET/*charset*/, 
									OUT_DEFAULT_PRECIS/*out precis*/,
									CLIP_DEFAULT_PRECIS/*clip precis*/,				
									DEFAULT_QUALITY/*quality*/,
									FF_MODERN/*pitch and family*/,			
									NULL/*typeface*/ );
			COLORREF crTextColorWas = GetTextColor( hdc );
			HFONT fntWas = (HFONT) SelectObject( hdc, fntState );

			//-- Drawing helpers.  precalc canvas center and size 
			int iX, iY;
			SIZE sizeText;			// buffer for text size calculation
			sizeText.cx=1;
			sizeText.cy=1;
			COLORREF crFill;
			HBRUSH brFill;
			RECT rct;

			//-- Draw Constraints
			HPEN penConHigh = CreatePen( PS_SOLID, 1, RGB(0,0,0) );
			HPEN penConLow = CreatePen( PS_SOLID, 1, RGB(64,64,255) );
			HPEN penConWas = (HPEN) SelectObject( hdc, penConHigh );

			CModelConstraint * pConstraint;
			for( int iConstraint=0; iConstraint<m_oModel.m_aConstraints.Count(); ++iConstraint ) { 
				pConstraint = m_oModel.Constraint(iConstraint);
				iX = XCanvas( pConstraint->m_dXPos );
				iY = YCanvas( pConstraint->m_dYPos );
				
				//----- Highlight selected Constraint 
				if( m_iCapType==CAP_CONSTRAINT && m_iCapItem==iConstraint ) {
					SelectObject( hdc, penConHigh );
					Rectangle( hdc, iX-CONSTRAINTWIDTH-2, iY-CONSTRAINTHEIGHT-2, 
									iX+CONSTRAINTWIDTH+3, iY+CONSTRAINTHEIGHT+3 );
					}
				else {
					SelectObject( hdc, penConLow );
					Rectangle( hdc, iX-CONSTRAINTWIDTH-1, iY-CONSTRAINTHEIGHT-1, 
									iX+CONSTRAINTWIDTH+2, iY+CONSTRAINTHEIGHT+2 );
					}

				//----- Box.  Note : Fillrect excludes right and bottom so +1 is added to compensate
				brFill = CreateSolidBrush( CR_CONSTRAINTBGR );
				SetRect( &rct, iX-CONSTRAINTWIDTH, iY-CONSTRAINTHEIGHT, iX+CONSTRAINTWIDTH+1, iY+CONSTRAINTHEIGHT+1 );
				
				FillRect( hdc, &rct, brFill );
				DeleteObject( brFill );
				
				//----- Text.
				pConstraint->Text( bufText );
				GetTextExtentPoint32( hdc, bufText, strlen(bufText), &sizeText );
				SetTextColor( hdc, CR_CONSTRAINTTEXT );
				ExtTextOut( hdc, iX-sizeText.cx/2, iY-sizeText.cy/2, 0, NULL, bufText, strlen(bufText), NULL );
				}

			SelectObject( hdc, penConWas );
			DeleteObject( penConHigh );
			DeleteObject( penConLow );

			SelectObject( hdc, fntWas );

			//-- Draw states 
			CModelState * pState;
			for( int iState=0; iState<m_oModel.m_aStates.Count(); ++iState ) { 
				pState = m_oModel.State(iState);
				iX = XCanvas( pState->m_dXPos );
				iY = YCanvas( pState->m_dYPos );
				
				//----- Highlight selected state 
				if( m_iMode==MODE_EDITSTATE && m_iCapType==CAP_STATE && m_iCapItem==iState ) 
					Rectangle( hdc, iX-STATERADIUS-2, iY-STATERADIUS-2, iX+STATERADIUS+3, iY+STATERADIUS+3 );

				//----- Box.  Note : Fillrect excludes right and bottom so +1 is added to compensate
				crFill = m_oModel.Class(pState->m_iClass)->m_cr;// StateColor( pState->m_iClass );
				brFill = CreateSolidBrush( crFill );
				SetRect( &rct, iX-STATERADIUS, iY-STATERADIUS, iX+STATERADIUS+1, iY+STATERADIUS+1 );
				FillRect( hdc, &rct, brFill );
				DeleteObject( brFill );
				
				//----- Text.  Color is black if sum of RGB components >= 384 
				sprintf( bufText, "%d", iState );
				GetTextExtentPoint32( hdc, bufText, strlen(bufText), &sizeText );
				SetTextColor( hdc, ((crFill&0xFF0000)>>16)+((crFill&0xFF00)>>8)+((crFill&0xFF))>=384 ? CR_STATEDARK : CR_STATELIGHT );
				ExtTextOut( hdc, iX-sizeText.cx/2, iY-sizeText.cy/2, 0, NULL, bufText, strlen(bufText), NULL );
				}
			SelectObject( hdc, fntWas );
			DeleteObject( fntState );
			
			//-- Draw Rates 
			SetTextColor( hdc, RGB(32,32,32) );
			HFONT fntRate;
			CModelState * pState0, * pState1;
			POINT pt0;				// connector start point
			POINT pt1;				// connector end point
			int iDY, iDX;			// Delta X, Y
			int iAdjustX, iAdjustY;	// amount to adjust x,y to get outside state box
			POINT ptA;				// arrow endpoint
			int iEscape;			// escapement angle (deci-degrees) for text drawn along rate line 
			double dAngle;			// escapement angle (radians) for text drawn along rate line 
			int itx, ity;
			
			for( int iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) { 
				CModelRate * pRate = m_oModel.Rate( iRate );

				// calc center of state 0 
				pState0 = m_oModel.State(pRate->m_aiStates[0]);
				pt0.x = XCanvas( pState0->m_dXPos );
				pt0.y = YCanvas( pState0->m_dYPos );
				
				// calc center of state 1 
				pState1 = m_oModel.State(pRate->m_aiStates[1]);
				pt1.x = XCanvas( pState1->m_dXPos );
				pt1.y = YCanvas( pState1->m_dYPos );
				
				//-- calc distance between and draw rate if not overlapping
				iDX = abs(pt1.x-pt0.x);
				iDY = abs(pt1.y-pt0.y);
				if( iDX > 2*STATERADIUS || iDY > 2*STATERADIUS ) { 
					//-- adjust x0, x1 to intersection points with state boxes					
					if( iDY > iDX ) { 
						iAdjustX = MulDiv(iDX,STATERADIUS,iDY) * (pt1.x>pt0.x?1:-1);
						iAdjustY = pt1.y>pt0.y?STATERADIUS:-1*STATERADIUS;
						}
					else { 
						iAdjustX = pt1.x>pt0.x?STATERADIUS:-1*STATERADIUS;
						iAdjustY = MulDiv(iDY,STATERADIUS,iDX) * (pt1.y>pt0.y?1:-1);
						}
					pt0.x += (iAdjustX+(iAdjustX>0?1:iAdjustX<0?-1:0));
					pt0.y += (iAdjustY+(iAdjustY>0?1:iAdjustY<0?-1:0));
					pt1.x -= iAdjustX;
					pt1.y -= iAdjustY;

					//-- Draw arrowhead, connector, arrowhead
					CalcArrowPoint( pt1, pt0, &ptA, true );
					MoveToEx( hdc, ptA.x, ptA.y, NULL );
					LineTo( hdc, pt0.x, pt0.y );
					LineTo( hdc, pt1.x, pt1.y );
					CalcArrowPoint( pt0, pt1, &ptA, true );
					LineTo( hdc, ptA.x, ptA.y );
					
					//-- draw text 
					if( pt1.x==pt0.x ) {
						dAngle=pt0.y>pt1.y?3*M_PI/2:M_PI/2;
						iEscape=pt0.y>pt1.y?2700:900; 
						}
					else {
						dAngle = -1.0 * atan( double(pt1.y-pt0.y) / double(pt1.x-pt0.x) );
						iEscape = int( 1800.0 * dAngle / M_PI );
						}
						
					fntRate = CreateFont( //TODO - doc suggests for MM_TEXT, nHeight = -MulDiv(PointSize, GetDeviceCaps(hDC, LOGPIXELSY), 72);
									-12 /*Height*/,		
									0 /*Width*/, 
									iEscape/*escapement*/,	
									0/*Orientation*/,  
									m_iCapType==CAP_RATE && m_iCapItem==iRate?FW_BOLD:FW_NORMAL/*Weight*/ /*Highlight selected rate*/,		
									FALSE/*Italic*/,
									FALSE/*Underline*/,	
									FALSE/*Strikeout*/,
									ANSI_CHARSET/*charset*/, 
									OUT_DEFAULT_PRECIS/*out precis*/,
									CLIP_DEFAULT_PRECIS/*clip precis*/,				
									ANTIALIASED_QUALITY/*DEFAULT_QUALITY/*quality*/,
									FF_MODERN/*FF_SWISS /*FF_ROMAN /*pitch and family*/,			
									NULL/*typeface*/ );
					SelectObject( hdc, fntRate );

					//-- format rate text dependent on magnitude 					
					RateText( bufText, pRate->m_adK0[0], pRate->m_adK1[0], pRate->m_abPDep[0], pRate->m_abQDep[0] );
					GetTextExtentPoint32( hdc, bufText, strlen(bufText), &sizeText );
					itx = int(cos(-1.0*dAngle)*sizeText.cx/2-sin(-1.0*dAngle)*sizeText.cy);
					ity = int(sin(-1.0*dAngle)*sizeText.cx/2+cos(-1.0*dAngle)*sizeText.cy);
					ExtTextOut( hdc, (pt0.x+pt1.x)/2-itx, (pt0.y+pt1.y)/2-ity, 0, NULL, bufText, strlen(bufText), NULL );

					RateText( bufText, pRate->m_adK0[1], pRate->m_adK1[1], pRate->m_abPDep[1], pRate->m_abQDep[1] );
					GetTextExtentPoint32( hdc, bufText, strlen(bufText), &sizeText );
					itx = int(cos(-1.0*dAngle)*sizeText.cx/2);
					ity = int(sin(-1.0*dAngle)*sizeText.cx/2);
					ExtTextOut( hdc, (pt0.x+pt1.x)/2-itx, (pt0.y+pt1.y)/2-ity, 0, NULL, bufText, strlen(bufText), NULL );

					SelectObject( hdc, fntWas );
					DeleteObject( fntRate );
					}
				}

			// remember size of rate text for validating clicks on rates
			// if no rates were drawn this is not valid, but there are no rates to check regardless.
			m_iRateTextHeight = sizeText.cy;
			
			//-- If constructing rate, draw the new rate 
			if( m_iMode == MODE_NEWRATE2 ) {
				HPEN penBlue = CreatePen( PS_SOLID, 4, RGB(0,0,255) );
				HPEN penWas = (HPEN) SelectObject( hdc, penBlue );
				pState = m_oModel.State(m_iRateStartState);
				MoveToEx( hdc, XCanvas(pState->m_dXPos), YCanvas(pState->m_dYPos), NULL );
				LineTo( hdc, m_iCapX, m_iCapY );
				SelectObject( hdc, penWas );
				DeleteObject( penBlue );
				}

			//-- If modifying the classes, show class buttons
			if( m_iMode == MODE_EDITCLASS ) { 
				SelectClipRgn( hdc, NULL );
				
				RECT rctClass;
				rctClass.left = m_iClLeft;
				rctClass.right = m_iClLeft+CLSEL_WIDTH;
				rctClass.top = m_iClTop;
				rctClass.bottom = m_iClTop + CLSEL_HEIGHT;
				HBRUSH brClass;
				int iRow=0; 
				COLORREF crFill;
				for( int iClass=0; iClass<MAXCLASSES && iClass<m_oModel.m_aClasses.Count(); ++iClass ) { 
					//-- Outline selected class
					if( m_iSelClass==iClass )
						Rectangle( hdc, rctClass.left-CLSEL_EXT, rctClass.top-CLSEL_EXT, 
										rctClass.right+CLSEL_EXT, rctClass.bottom+CLSEL_EXT );
						
					//-- filled box
					//crFill = StateColor(iClass);
					crFill = m_oModel.Class(iClass)->m_cr;
					brClass = CreateSolidBrush( crFill );
					FillRect( hdc, &rctClass, brClass );
					DeleteObject( brClass );

					//-- Text
					sprintf( bufText, "%d", iClass );
					GetTextExtentPoint32( hdc, bufText, strlen(bufText), &sizeText );
					SetTextColor( hdc, ((crFill&0xFF0000)>>16)+((crFill&0xFF00)>>8)+((crFill&0xFF))>=384 ? CR_STATEDARK : CR_STATELIGHT );
					ExtTextOut( hdc, (rctClass.left+rctClass.right-sizeText.cx)/2, (rctClass.bottom+rctClass.top-sizeText.cy)/2, 0, NULL, bufText, strlen(bufText), NULL );

					//-- Move to next box
					if( (iClass+1)%CLSEL_PERROW==0 ) { 
						rctClass.top += CLSEL_DELTAY;
						rctClass.bottom += CLSEL_DELTAY;
						rctClass.left = m_iClLeft;
						rctClass.right = m_iClLeft+CLSEL_WIDTH;
						}
					else { 
						rctClass.left += CLSEL_DELTAX;
						rctClass.right += CLSEL_DELTAX;
						}
					}
				}	
				
			//-- Restore dc state
			SelectObject( hdc, fntWas );
			SetTextColor( hdc, crTextColorWas );
			SetBkMode( hdc, iBkModeWas );

			EndPaint(m_hDlg, &ps);
			}			

		//=========================================================== Mouse capture handlers

		//-- check if x,y coordinate corresponds to an existing state
		// returns state index or -1 if no state clicked.
		int StateFromPoint( int iX, int iY ) { 
			int iXState, iYState;
			for( int iState=0; iState<m_oModel.m_aStates.Count(); ++iState ) { 
				CModelState * pState = m_oModel.State(iState);
				iXState = XCanvas( pState->m_dXPos );
				iYState = YCanvas( pState->m_dYPos );
				if( iX>=iXState-STATERADIUS && iY>=iYState-STATERADIUS
						&& iX<=iXState+STATERADIUS+1 && iY<=iYState+STATERADIUS+1 )
					return iState;
					
				}
			return -1;
			}

		//-- check if x,y coordinate corresponds to an existing rate
		// returns rate index or -1 if no rate clicked.
		int RateFromPoint( int iX, int iY ) { 
			POINT pt0;				// connector start point
			POINT pt1;				// connector end point
			CModelState * pState0, * pState1;
			CModelRate * pRate;
			
			for( int iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) { 
				pRate = m_oModel.Rate( iRate );

				// calc center of state 0 
				pState0 = m_oModel.State( pRate->m_aiStates[0] );
				pt0.x = XCanvas( pState0->m_dXPos );
				pt0.y = YCanvas( pState0->m_dYPos );
				
				// calc center of state 1 
				pState1 = m_oModel.State( pRate->m_aiStates[1] );
				pt1.x = XCanvas( pState1->m_dXPos );
				pt1.y = YCanvas( pState1->m_dYPos );
				
				if( pRate->IsHit( iX, iY, pt0, pt1, m_iRateTextHeight, STATERADIUS ) != 0 )
					return iRate;
				}
				
			return -1;
			}

		//-- check if x,y coordinate corresponds to an existing Constraint
		// returns index or -1 if no constraint clicked.
		int ConstraintFromPoint( int iX, int iY ) { 
			int iXCon, iYCon;
			for( int i=0; i<m_oModel.m_aConstraints.Count(); ++i ) { 
				CModelConstraint * pConstraint = m_oModel.Constraint(i);
				iXCon = XCanvas( pConstraint->m_dXPos );
				iYCon = YCanvas( pConstraint->m_dYPos );
				if( iX>=iXCon-CONSTRAINTWIDTH && iY>=iYCon-CONSTRAINTHEIGHT
						&& iX<=iXCon+CONSTRAINTWIDTH+1 && iY<=iYCon+CONSTRAINTHEIGHT+1 )
					return i;
					
				}
			return -1;
			}

		//-- check if two specified states have a rate connecting them.
		int RateFromStates( int iState0, int iState1 ) { 			
			int * ai;
			for( int iRate=0; iRate<m_oModel.m_aRates.Count(); ++iRate ) { 
				ai = m_oModel.Rate( iRate )->m_aiStates;
				if( (ai[0]==iState0 && ai[1]==iState1) || (ai[0]==iState1 && ai[1]==iState0) ) 
					return iRate;
				}
				
			return -1;
			}

		bool StartRateDrag( int iX, int iY ) { 
			int iState = StateFromPoint( iX, iY );
			if( iState == -1 ) 
				return false;
				
			m_iRateStartState = iState;
			SetMode( MODE_NEWRATE2 );
			m_iCapX=max(min(m_rctCanvas.right, iX), m_rctCanvas.left );
			m_iCapY=max(min(m_rctCanvas.bottom, iY), m_rctCanvas.top );

			m_iCapType=CAP_CONSTRUCTRATE;
			SetCapture( m_hDlg );
			m_pCapture=(CDlgExtra *)(m_oModel.m_aStates[m_iRateStartState]);
			return true;
			}
		
		void StateAdd( int iX, int iY ) { 
			m_pCapture=NULL;			  //-- Inserting a new state
			m_oModel.AddState( 0, XPosition(iX), YPosition(iY) );
			SetUpdated();
			OnSelectState( iX, iY, m_oModel.m_aStates.Count()-1, (CModelState *)m_oModel.m_aStates.Last() );
			m_selClass.SetCheck( false );
			m_iNewSel=-1;
			}
		
			
		bool RMouseDown( short iX, short iY ) { 
			//if( m_iMode == MODE_VIEW ) { 
				if( StartRateDrag( iX, iY ) )  //-- dragging a new rate 
					return true;
				StateAdd( iX, iY );		// adding a new state 
			//	}
			return true;
			}
		
		bool LMouseDown( short iX, short iY ) { 
			m_pCapture=NULL;

			//-- Inserting a new state
			if( m_iMode == MODE_NEWSTATE ) { 
				StateAdd( iX, iY );
				return false;
				}
				
			if( m_iMode == MODE_NEWRATE1 ) { 
				if( ! StartRateDrag( iX, iY ) ) { 
					m_iNewSel=-1;
					m_selRate.SetCheck( false );
					SetMode( MODE_VIEW );
					return false;
					}

				return true;
				}

			//-- Inserting a new constraint
			if( m_iMode == MODE_NEWCONSTRAINT ) { 
				int aiStates[2]={0,1};
				m_oModel.AddConstraint( CONSTRAINT_FIXRATE, 2, aiStates, XPosition(iX), YPosition(iY) );
				SetUpdated();
				OnSelectConstraint( iX, iY, m_oModel.m_aConstraints.Count()-1, (CModelConstraint *)m_oModel.m_aConstraints.Last() );
				m_selConstraint.SetCheck( false );
				m_iNewSel=-1;
				return false;
				}

			//-- selecting a new class during class properties edit 
			if( m_iMode == MODE_EDITCLASS && iX>=m_iClLeft &&iY>=m_iClTop 
					&&iX<=m_iClLeft+CLSEL_PERROW*CLSEL_DELTAX+CLSEL_WIDTH
					&&iY<=m_iClTop+CLSEL_DELTAY*3 ) { 
				int iRow = min((iY-m_iClTop)/CLSEL_DELTAY,2);
				int iCol = min((iX-m_iClLeft)/CLSEL_DELTAX,CLSEL_PERROW-1);
				int iSelClass = iRow*CLSEL_PERROW+iCol;
				if( iSelClass<m_oModel.m_aClasses.Count() && iSelClass<MAXCLASSES && iSelClass!=m_iSelClass )
					SetSelClass(iSelClass);
				}
				

			//-- check if clicked on state 
			int iClickedState = StateFromPoint( iX, iY );
			if( iClickedState != -1 ) { 
				CModelState * pState = m_oModel.State(iClickedState);
				OnSelectState( iX, iY, iClickedState, pState );
				m_pCapture=(CDlgExtra *)pState;
				SetCapture( m_hDlg );
				return true;
				}
			
			//-- check if clicked on rate 
			int iClickedRate = RateFromPoint( iX, iY );
			if( iClickedRate != -1 ) { 
				OnSelectRate( iX, iY, iClickedRate, m_oModel.Rate( iClickedRate ) );
				return true;
				}
			
			//-- check if clicked on constraint
			int iClickedConstraint = ConstraintFromPoint( iX, iY );
			if( iClickedConstraint != -1 ) { 
				CModelConstraint * pConstraint = m_oModel.Constraint(iClickedConstraint);
				OnSelectConstraint( iX, iY, iClickedConstraint, pConstraint );
				m_pCapture=(CDlgExtra *)pConstraint;
				SetCapture( m_hDlg );
				return true;
				}

			CanvasInvalidate();
			if( m_rctCanvas.left<iX && m_rctCanvas.right>iX && m_rctCanvas.top<iY && m_rctCanvas.bottom>iY ) { 
				SetMode( MODE_VIEW );
				m_pCapture=NULL;
				m_iCapItem=-1;
				}
			return false;	
			}
			
		void MouseMove( short iX, short iY ){
			//-- if dragging state 
			if( m_iCapType==CAP_STATE ) { 
				((CModelState *)m_pCapture)->m_dXPos = m_dCapX + XPosition( iX ) - XPosition( m_iCapX );
				((CModelState *)m_pCapture)->m_dYPos = m_dCapY + YPosition( iY ) - YPosition( m_iCapY );
				((CModelState *)m_pCapture)->m_dXPos = 3 * int( ((CModelState *)m_pCapture)->m_dXPos /3 );
				((CModelState *)m_pCapture)->m_dYPos = 3 * int( ((CModelState *)m_pCapture)->m_dYPos /3 );
				SetUpdated();
				CanvasInvalidate();
				}

			//-- if constructing new rate 
			if( m_iCapType==CAP_CONSTRUCTRATE ) { 
				m_iCapX=max(min(m_rctCanvas.right, iX), m_rctCanvas.left );
				m_iCapY=max(min(m_rctCanvas.bottom, iY), m_rctCanvas.top );
				CanvasInvalidate();
				}

			//-- if dragging constraint
			if( m_iCapType==CAP_CONSTRAINT ) { 
				((CModelConstraint *)m_pCapture)->m_dXPos = m_dCapX + XPosition( iX ) - XPosition( m_iCapX );
				((CModelConstraint *)m_pCapture)->m_dYPos = m_dCapY + YPosition( iY ) - YPosition( m_iCapY );
				SetUpdated();
				CanvasInvalidate();
				}
			}

		void MouseUp( short iX, short iY ){
			ReleaseCapture();
			m_pCapture=NULL;
			CanvasInvalidate();

			if( m_iMode == MODE_NEWRATE2 ) { 
				m_iNewSel=-1;
				m_selRate.SetCheck( false );
				int iClickedState = StateFromPoint( iX, iY );
				if( iClickedState==-1 || iClickedState==m_iRateStartState
						|| -1!=RateFromStates( m_iRateStartState, iClickedState ) ) { 
					SetMode( MODE_VIEW );
					}
				else { 
					m_oModel.AddRate( m_iRateStartState, iClickedState );	
					OnSelectRate( iX, iY, m_oModel.m_aRates.Count()-1, (CModelRate *)m_oModel.m_aRates.Last() );
					SetUpdated();
					}
				}

			}

		//=========================================================== Message handlers
		void OnNewModel( ) { 
			SetFileName( "New Model" );
			m_oModel.NewModel( );
			SetClassList();
			//m_bNewFile=true;
			}
			
		void OnSetModel( CModel * pModel ) { 
			m_oModel.ClearModel();
			m_oModel.Copy( *pModel );
			SetClassList();
			}
			
		//--
		void OnLoadFile( const char * pchFile ) { 
			SetFileName( pchFile );
			m_bUpdated=false;
			if( ! m_oModel.OpenQmf( pchFile ) ) { 
				MessageBox( m_hDlg, "Error accessing file", "Open Model file", MB_OK );
				ModelMenuOff();
				DestroyWindow( GetParent( m_hDlg ) );
				return;
				} 
			
			SetClassList();
			}			

		//-- Open MIL dialogue to analyze data
		void OnMil( ) { 
			OpenVwMil( m_hDlg, &m_oModel );
			}

		//-- Open dialogue to Generate simulated data
		void OnSimulate( ) { 
			OpenVwSimulate( m_hDlg, &m_oModel );
			}

		//-- Open Statistics window ( similar to model report ) 
		void OnStatistics( ) { 
			CStringArray ac;
			ac.AddString( "Export PDF" );
			ac.AddString( "Save PDF" );

			CModelRpt * poRpt = new CModelRpt( &m_oModel );
			poRpt->WinOpen( "Model Report", false, &ac );
			}

		//--
		bool OnSave( ) { 
			char bufName[_MAX_PATH];
			strcpy( bufName, m_cFileName.Buffer() );
			m_bUpdated=false;
			if( ! GetSaveFileName( 
							bufName, 
							"Model file (*.qmf)\0*.qmf\0All Files (*.*)\0*.*\0\0",
							"qmf", 
							GetParent(m_hDlg)))
				return false;
				
			SetFileName(bufName);
			m_oModel.WriteQmf( bufName );
			UpdateModelMRU( bufName );
			return true;
			}			

		//--
		void OnExit() { 
			if( Updated() && IDYES==MessageBox(GetParent(m_hDlg),"Save changes to model ?", "Model Close", MB_YESNO ) )  
				OnSave();
			}

		//--
		void OnBtnClose( ) { 
			ModelMenuOff();
			DestroyWindow( GetParent( m_hDlg ) );
			}			

		//--
		void OnBtnDelete( ) { 
			if( m_iMode!=MODE_EDITRATE && m_iMode!=MODE_EDITCONSTRAINT 
					&& m_iMode!=MODE_EDITSTATE && m_iMode!=MODE_EDITCLASS ) 
				return;
				
			if( IDNO==MessageBox( m_hDlg, 
						m_iMode==MODE_EDITRATE 
						? "Delete the selected rate" 
						: m_iMode==MODE_EDITSTATE 
							? "Delete the selected state"
							: m_iMode==MODE_EDITCLASS
								? "Delete the selected class"
								: "Delete the selected constraint", 
							 "Confirm delete", MB_YESNO ) ) 
				return;
				
			if( m_iMode==MODE_EDITRATE ) { 
				m_oModel.DeleteRate( m_iCapItem );
				SetUpdated();
				CanvasInvalidate();				
				SetMode( MODE_VIEW );
				}

			else if( m_iMode==MODE_EDITSTATE ) { 
				m_oModel.DeleteState( m_iCapItem );
				SetMode( MODE_VIEW );
				SetUpdated();
				CanvasInvalidate();				
				SetMode( MODE_VIEW );
				}

			else if( m_iMode==MODE_EDITCONSTRAINT ) { 
				m_oModel.DeleteConstraint( m_iCapItem );
				SetUpdated();
				CanvasInvalidate();				
				SetMode( MODE_VIEW );
				}

			else if( m_iMode==MODE_EDITCLASS ) { 
				m_oModel.DeleteClass( m_iSelClass );
				SetUpdated();
				PropertiesInvalidate();				
				CanvasInvalidate();	
				if( m_iSelClass >= m_oModel.m_aClasses.Count() ) 
					m_iSelClass=0;
				//SetMode( MODE_VIEW );
				}
			}			
	
		//-- open energy landscape view 
		void OnEnergyLandscape() {
			PostMessage( GetParent(GetClientWin()), WM_COMMAND, ID_ENERGYMODEL_VIEW, (LPARAM) &m_oModel );
			}

	
		//-- Edit the class table
		void OnBtnClasses( ) { 
			SetSelClass( 0 );
			}

		//--
		void OnBtnUndo( ) { 
			// TODO 
			}			

		//--  Selection of current "new" item (state,rate,constraint)
		void OnSelNewItem( int iItem ) { 
			m_iCapType=CAP_NONE;
			CanvasInvalidate();
			if( m_iNewSel!=iItem ) {
				m_iNewSel=iItem;
				if( iItem==1 ) 
					SetMode( MODE_NEWSTATE );
				else if( iItem==2 ) 
					SetMode( MODE_NEWRATE1 );
				else if( iItem==3 ) 
					SetMode( MODE_NEWCONSTRAINT );
				}
			else { 
				m_iNewSel=-1;
				m_selClass.SetCheck( false );
				m_selRate.SetCheck( false );
				m_selConstraint.SetCheck( false );
				SetMode(MODE_VIEW);
				}
			}			

		//-- selection of new class to edit 
		void SetSelClass( int iSelClass ) {
			m_iMode=MODE_VIEW;				// prevent update messages while setting controls.

			m_iSelClass=iSelClass;
			PropertiesInvalidate();
			CModelClass * pClass = m_oModel.Class(iSelClass);
			m_edtClassAmp.SetTextDouble( pClass->m_dAmp );
			m_edtClassStd.SetTextDouble( pClass->m_dStdDev );
			m_edtClassCond.SetTextDouble( pClass->m_dConditioning );

			SetMode( MODE_EDITCLASS );
			}

		//-- Check if user wants to save file before closing window.  false if user cancels save.		
		bool Close( ) { 
			return OnSave(); 
			}	

		//----- Check if a window can be closed.
		bool CanClose( ) { 
			return ! Updated() || m_oModel.GetStatus()==0; 
			}	

		//-- Limit display size to at least 690 x 275
		void OnMinMaxInfo( MINMAXINFO * p ) {  
			p->ptMinTrackSize.x = 690;
			p->ptMinTrackSize.y = 275;
			}

		//-- related to window sizing 
		void OnSize( WPARAM wParam, LPARAM * p ) { 
			/*	In response to a resizing, which includes the initial sizing of the window, 
				this function resizes the display canvas.
			*/
			SetMode( MODE_VIEW );
			m_pCapture=NULL;
			m_iCapItem=-1;
			
			//-- Desired size
			int iDlgWid=LOWORD(*p);
			int iDlgHgt=HIWORD(*p);

			//-- dialogue window info 
			RECT rcPar;
			GetBounds( &rcPar );
			MoveWindow( m_hDlg, 0, 0, iDlgWid, iDlgHgt, TRUE );
			
			//-- Resize picture control 
			m_picCanvas.GetControlBounds( &m_rctCanvas );
			m_rctCanvas.left--;							
			m_rctCanvas.top--;							
			m_rctCanvas.right = iDlgWid-(rcPar.right-m_rctCanvas.right)+1;		// indent right same as left
			m_rctCanvas.bottom = iDlgHgt-(rcPar.bottom-m_rctCanvas.bottom);	// indent bottom same at left 
			MoveWindow( m_picCanvas.GetHandle(), 
							m_rctCanvas.left, 
							m_rctCanvas.top, 
							m_rctCanvas.right-m_rctCanvas.left,
							m_rctCanvas.bottom-m_rctCanvas.top+1,	
							TRUE );

			m_iXSize = (m_rctCanvas.right-m_rctCanvas.left);
			m_iYSize = (m_rctCanvas.bottom-m_rctCanvas.top);
			m_iXMid = (m_rctCanvas.right+m_rctCanvas.left)/2;			
			m_iYMid = (m_rctCanvas.bottom+m_rctCanvas.top)/2;			
			}			

		//---------- Mouse drag support 
		void OnSelectState( int iX, int iY, int iState, CModelState * pState ) { 
			//-- if already selected, exit.
			if( m_iCapType==CAP_STATE && m_iCapItem==iState ) 
				return;
				
			m_iCapType=CAP_STATE;
			m_iCapItem=iState;
			m_iCapX=iX;
			m_iCapY=iY;
			m_dCapX=pState->m_dXPos;
			m_dCapY=pState->m_dYPos;
			CanvasInvalidate();

			int iSel = m_cmbStateClass.FindStringInt( pState->m_iClass, "%d" );
			if( iSel==CB_ERR )	
				SetMode(MODE_VIEW);	// this should never occur...
			else { 
				m_iMode=MODE_VIEW;	//-- feign view mode so setting controls update messages do not get handled
				m_cmbStateClass.SetCurSel( iSel );
				m_edtStateStartProb.SetTextDouble( pState->m_dStartProb );
				m_edtStateAmp.SetTextDouble( m_oModel.Class(pState->m_iClass)->m_dAmp );
				m_edtStateStd.SetTextDouble( m_oModel.Class(pState->m_iClass)->m_dStdDev );
				
				char buf[64];
				sprintf( buf, "State [%d]", m_iCapItem );
				m_lblState.SetText( buf );
				
				SetMode( MODE_EDITSTATE );
				}
			}

		void OnSelectRate( int iX, int iY, int iRate, CModelRate * pRate ) {
			//-- if already selected, exit.
			if( m_iCapType==CAP_RATE && m_iCapItem==iRate ) 
				return;
				
			m_iCapType=CAP_RATE;
			m_iCapItem=iRate;
			m_iCapX=iX;
			m_iCapY=iY;
			CanvasInvalidate();

			m_iMode=MODE_VIEW;	//-- feign view mode so setting controls update messages do not get handled
			
			char buf[64];
			sprintf( buf, "%d %d", pRate->m_aiStates[0], pRate->m_aiStates[1] );
			m_lblRateFwd.SetText( buf );
			sprintf( buf, "%d %d", pRate->m_aiStates[1], pRate->m_aiStates[0] );
			m_lblRateBwd.SetText( buf );
			
			m_edtRatePF.SetCheck( pRate->m_abPDep[0] );
			m_edtRateQF.SetCheck( pRate->m_abQDep[0] );
			m_edtRatePB.SetCheck( pRate->m_abPDep[1] );
			m_edtRateQB.SetCheck( pRate->m_abQDep[1] );
			
			EnableRateNames();

			m_edtRateK0F.SetTextDouble(pRate->m_adK0[0]);
			m_edtRatePNameF.SetText(pRate->m_acPName[0].Buffer());
			m_edtRateK1F.SetTextDouble(pRate->m_adK1[0]);
			m_edtRateQNameF.SetText(pRate->m_acQName[0].Buffer());
			m_edtRateK0B.SetTextDouble(pRate->m_adK0[1]);
			m_edtRatePNameB.SetText(pRate->m_acPName[1].Buffer());
			m_edtRateK1B.SetTextDouble(pRate->m_adK1[1]);
			m_edtRateQNameB.SetText(pRate->m_acQName[1].Buffer());
			SetMode( MODE_EDITRATE );
			}

		void OnSelectConstraint( int iX, int iY, int iConstraint, CModelConstraint * pConstraint ) { 
			char buf[64];
			//-- if already selected, exit.
			if( m_iCapType==CAP_CONSTRAINT && m_iCapItem==iConstraint ) 
				return;
				
			//-- fill in rate list combos
			m_cmbCnsRate.ClearContents();
			m_cmbCnsRefRate.ClearContents();
			int iRate, iRates=m_oModel.m_aRates.Count();
			for( iRate=0; iRate<iRates; ++iRate ) { 
				CModelRate * pRate = m_oModel.Rate(iRate);

				sprintf( buf, "%d->%d", pRate->m_aiStates[0], pRate->m_aiStates[1] );
				m_cmbCnsRate.AddString(buf);
				m_cmbCnsRefRate.AddString(buf);

				sprintf( buf, "%d->%d", pRate->m_aiStates[1], pRate->m_aiStates[0] );
				m_cmbCnsRate.AddString(buf);
				m_cmbCnsRefRate.AddString(buf);
				}
				
			m_iCapType=CAP_CONSTRAINT;
			m_iCapItem=iConstraint;
			m_iCapX=iX;
			m_iCapY=iY;
			m_dCapX=pConstraint->m_dXPos;
			m_dCapY=pConstraint->m_dYPos;
			CanvasInvalidate();

			m_iMode=MODE_VIEW;	//-- feign view mode so setting controls update messages do not get handled
			
			//-- Select constraint type in combo
			if( pConstraint->m_iType==CONSTRAINT_FIXRATE ) 
				m_cmbCnsType.SetCurSel( sCONS_FIXRATE );
			else if( pConstraint->m_iType==CONSTRAINT_FIXEXP ) 
				m_cmbCnsType.SetCurSel( sCONS_FIXEXP );
			else if( pConstraint->m_iType==CONSTRAINT_SCALERATE ) 
				m_cmbCnsType.SetCurSel( sCONS_SCALERATE );
			else if( pConstraint->m_iType==CONSTRAINT_SCALEEXP ) 
				m_cmbCnsType.SetCurSel( sCONS_SCALEEXP );
			else { 
				pConstraint->m_iType=CONSTRAINT_FIXRATE; 
				m_cmbCnsType.SetCurSel( sCONS_FIXRATE );
				}		
			
			//-- select constraint rates in lists 
			sprintf( buf, "%d->%d", pConstraint->m_aiStates[0], pConstraint->m_aiStates[1] );
			iRate = m_cmbCnsRate.FindString(buf);
			if( iRate>=0 ) 
				m_cmbCnsRate.SetCurSel(iRate);
				
			sprintf( buf, "%d->%d", pConstraint->m_aiStates[2], pConstraint->m_aiStates[3] );
			iRate = m_cmbCnsRefRate.FindString(buf);
			if( iRate>=0 ) 
				m_cmbCnsRefRate.SetCurSel(iRate);

			CanvasInvalidate();
			SetMode( MODE_EDITCONSTRAINT );
			}

		//-- Notification that the state properties have changed
		void OnChangeStateClass() { 
			if( m_iMode == MODE_EDITSTATE ) {
				int iClass = m_cmbStateClass.GetTextInt();
				((CModelState *)(m_oModel.m_aStates[m_iCapItem]))->m_iClass = iClass;
				CanvasInvalidate();

				m_edtStateAmp.SetTextDouble( m_oModel.Class(iClass)->m_dAmp );
				m_edtStateStd.SetTextDouble( m_oModel.Class(iClass)->m_dStdDev );
				SetUpdated();
				}
			}

		void OnChangeStateProbability() { 
			if( m_iMode == MODE_EDITSTATE ) {
				((CModelState *)(m_oModel.m_aStates[m_iCapItem]))->m_dStartProb = m_edtStateStartProb.GetTextDouble();
				SetUpdated();
				}
			}

		//-- Notification that the constraint properties have changed
		void OnChangeConstraint( int iCmb ) { // iCmb=0:type 1:rate 2:ref rate
			if( m_iMode==MODE_EDITCONSTRAINT ) {
				CModelConstraint * pCon = ((CModelConstraint *)(m_oModel.m_aConstraints[m_iCapItem]));
				
				//-- extract state numbers from combo's
				CString c;
				int ai[4]={-1,-1,-1,-1};
				m_cmbCnsRate.GetText( &c );
				char * pchPos = strstr( c.ConstBuffer(), "->" );
				if( pchPos != NULL ) { 
					ai[0]=atoi(c.Left(pchPos-c.ConstBuffer()).ConstBuffer());
					ai[1]=atoi(pchPos+2);
					}
				m_cmbCnsRefRate.GetText( &c );
				pchPos = strstr( c.ConstBuffer(), "->" );
				if( pchPos != NULL ) { 
					ai[2]=atoi(c.Left(pchPos-c.ConstBuffer()).ConstBuffer());
					ai[3]=atoi(pchPos+2);
					}

				if( iCmb==0 ) {				// change to constraint type
					int iType=CONSTRAINT_FIXRATE;
					int iSel=m_cmbCnsType.GetCurSel();
					int iStates=2;
					if( iSel==0 ) 
						iType=CONSTRAINT_FIXRATE;
					else if( iSel==1 ) 
						iType=CONSTRAINT_FIXEXP;
					else if( iSel==2 ) {
						iType=CONSTRAINT_SCALERATE;
						iStates=4;
						}
					else if( iSel==3 ) {
						iType=CONSTRAINT_SCALEEXP;
						iStates=4;
						}
					pCon->m_iType=iType;
					pCon->SetStates( iStates, ai[0], ai[1], ai[2], ai[3] );	
					}
				else if( iCmb==1 && pCon->m_iStates>=2 ) {		// change to rate 
					pCon->m_aiStates[0]=ai[0];
					pCon->m_aiStates[1]=ai[1];
					}
				else if( iCmb==2 && pCon->m_iStates>=4 ) {		// change to reference rate 
					pCon->m_aiStates[2]=ai[2];
					pCon->m_aiStates[3]=ai[3];
					}
				
				SetUpdated();
				CanvasInvalidate();
				}
			}

		//-- Notification that the rate properties have changed
		void OnChangeRate() { 
			if( m_iMode != MODE_EDITRATE ) 
				return;
				
			CModelRate * pRate = (CModelRate *)(m_oModel.m_aRates[m_iCapItem]);
			pRate->m_adK0[0] = m_edtRateK0F.GetTextDouble();
			pRate->m_abPDep[0] = m_edtRatePF.GetCheck();
			m_edtRatePNameF.GetText( &pRate->m_acPName[0] );

			pRate->m_adK1[0] = m_edtRateK1F.GetTextDouble();
			pRate->m_abQDep[0] = m_edtRateQF.GetCheck();
			m_edtRateQNameF.GetText( &pRate->m_acQName[0] );

			pRate->m_adK0[1] = m_edtRateK0B.GetTextDouble();
			pRate->m_abPDep[1] = m_edtRatePB.GetCheck();
			m_edtRatePNameB.GetText( &pRate->m_acPName[1] );

			pRate->m_adK1[1] = m_edtRateK1B.GetTextDouble();
			pRate->m_abQDep[1] = m_edtRateQB.GetCheck();
			m_edtRateQNameB.GetText( &pRate->m_acQName[1] );

			SetUpdated();
			CanvasInvalidate();
			}
			
		//-- Notification for the class properties 			
		void OnChangeClass() { 
			if( m_iMode != MODE_EDITCLASS ) 
				return;
			CModelClass * pClass = m_oModel.Class( m_iSelClass );
			pClass->m_dAmp = m_edtClassAmp.GetTextDouble();
			pClass->m_dStdDev = m_edtClassStd.GetTextDouble();
			SetUpdated();
			pClass->m_dConditioning = m_edtClassCond.GetTextDouble();
			}

		//-- Handler for button to change class color within this model
		void OnChangeClassColor() {
			COLORREF acrTemp[16];
			memcpy( acrTemp, m_acrTraceColor, 16*sizeof(COLORREF) );
			
			CHOOSECOLOR cc;

			ZeroMemory(&cc, sizeof(cc));
			cc.lStructSize = sizeof(cc);
			cc.hwndOwner = m_hDlg;
			cc.lpCustColors = acrTemp; //NULL;	
			cc.rgbResult = m_oModel.Class( m_iSelClass )->m_cr;
			cc.Flags = CC_FULLOPEN | CC_RGBINIT;	// CC_FULLOPEN CC_ANYCOLOR | 

			if( ChooseColor(&cc) ) 
				m_oModel.Class( m_iSelClass )->m_cr=cc.rgbResult;
			SetUpdated();
			InvalidateRect( m_hDlg, NULL, false );
			}
			
		//-- add a class to the model class array if not full
		void OnAddClass() { 
			m_oModel.AddClass();
			SetSelClass( m_oModel.m_aClasses.Count()-1 );			
			SetClassList();
			SetUpdated();
			}
			
		void OnModelChanged() {		
			SetUpdated();
			PropertiesInvalidate();				
			CanvasInvalidate();	
			}
			
		void OnMDIActivate( HWND hwndLosing, HWND hwndGaining ) {
			if( GetParent(m_hDlg)==hwndLosing )
				ModelMenuOff();
			if( GetParent(m_hDlg)==hwndGaining ) 
				ModelMenuOn();
			}

		//-- For painting the class edit button 
		HBRUSH CreateButtonBrush() { 
			if( hbrButton != NULL ) 
				DeleteObject(hbrButton);
			return hbrButton = CreateSolidBrush( m_oModel.Class(m_iSelClass)->m_cr );
			}
	};


//-----------------------------------------------------------------------------
BOOL CALLBACK ModelViewDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam) { 
	CVwModel * pVw;

	//-- create the CVwModel to go with the dialog 
	if( iMsg == WM_INITDIALOG ) { 
		pVw = new CVwModel;
		pVw->m_hDlg=hDlg;
		pVw->InitializeControls();
		pVw->OnMDIActivate( NULL, GetParent(hDlg) );	//		ModelMenuOn();
		SetWindowLong( GetParent(hDlg), 0, (long) pVw ) ;
		return TRUE;
		}

	//-- get the ViewClass, possibly NULL for initial messages 
	pVw = (CVwModel *) GetWindowLong(  GetParent( hDlg ), 0 );
	
	//-- Message handlers 
	int wmId, wmEvent;
	switch( iMsg ) { 
		case WM_MODEL_SETMODEL : 
			pVw->OnSetModel( (CModel *)lParam );
			return TRUE;
			
		case WM_MODEL_NEWFILE : 
			pVw->OnNewModel();
			return TRUE;
			
		case WM_MODEL_LOADFILE :
			pVw->OnLoadFile( (char *) lParam );
			return TRUE;

		case WM_CTLCOLORBTN :
			wmId = GetDlgCtrlID( (HWND)lParam );
			if( wmId==IDC_BTNCLASSCOLOR )
				return (BOOL) pVw->CreateButtonBrush();
			return FALSE;

		case WM_PAINT :
			pVw->OnPaint();
			return FALSE;

		case WM_MODEL_CHANGED : 
			pVw->OnModelChanged();
			if( wParam ) 
				PostMessage( GetClientWin(), WM_MDIACTIVATE, (WPARAM)GetParent(pVw->m_hDlg), 0 );
			return true;

		//-----
		case WM_COMMAND :
			wmId = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			switch( wmId ) { 
				//---- 'New' item mode
				case IDC_SEL_CLASS :
					pVw->OnSelNewItem( 1 );
					return TRUE;
					
				case IDC_SEL_RATE :
					pVw->OnSelNewItem( 2 );
					return TRUE;

				case IDC_SEL_CONSTRAINT :
					pVw->OnSelNewItem( 3 );
					return TRUE;

				//-- change to State Properties
				case IDC_CMBSTATECLASS :			//-- {Config} Combo Box
					if( wmEvent==CBN_SELCHANGE )
						pVw->OnChangeStateClass();
					break;
				case IDC_EDTSTATESTARTPROB :			
					if( wmEvent==EN_CHANGE )
						pVw->OnChangeStateProbability();
					break;
					
				//-- change to constraint properties 
				case IDC_CMB_CNS_TYPE :
					if( wmEvent==CBN_SELCHANGE )
						pVw->OnChangeConstraint(0);
					break;
				case IDC_CMB_CNS_RATE :
					if( wmEvent==CBN_SELCHANGE )
						pVw->OnChangeConstraint(1);
					break;
				case IDC_CMB_CNS_REFRATE :
					if( wmEvent==CBN_SELCHANGE )
						pVw->OnChangeConstraint(2);
					break;

				//-- Change to Rate Properties
				case IDC_EDTRATEK0F :
				case IDC_EDTRATEPNAMEF :
				case IDC_EDTRATEK1F :
				case IDC_EDTRATEQNAMEF :
				case IDC_EDTRATEK0B :
				case IDC_EDTRATEPNAMEB :
				case IDC_EDTRATEK1B :
				case IDC_EDTRATEQNAMEB :
					if( wmEvent==EN_CHANGE )
						pVw->OnChangeRate();
					break;

				case IDC_CHKRATEPF : //-- P and Q checkboxes cause enable / disable of name 
				case IDC_CHKRATEQF : 
				case IDC_CHKRATEPB : 
				case IDC_CHKRATEQB : 
					if( wmEvent==BN_CLICKED ){
						pVw->OnChangeRate();
						pVw->EnableRateNames();
						}
					break;

				//-- change to class properties
				case IDC_EDTCLASSAMP :
				case IDC_EDTCLASSSTD :
				case IDC_EDTCLASSCOND :
					if( wmEvent==EN_CHANGE )
						pVw->OnChangeClass();
					break;
				case IDC_BTNCLASSCOLOR :
					if( wmEvent==BN_CLICKED )
						pVw->OnChangeClassColor();
					break;
				case IDC_BTN_ADDCLASS :
					if( wmEvent==BN_CLICKED )
						pVw->OnAddClass();
					break;

				//-- action buttons
				case ID_MODEL_MIL :
					pVw->OnMil();
					return TRUE;
				
				case ID_MODEL_SIMULATE :
					pVw->OnSimulate();
					return TRUE;
				
				case ID_MODEL_STATISTICS :
					pVw->OnStatistics();
					return TRUE;
				
				case IDC_BTNCLOSE : 
					pVw->OnBtnClose();
					return TRUE;
					
				case IDC_BTNDELETE : 
					pVw->OnBtnDelete();
					return TRUE;
					
				case IDC_BTNENERGY :
					pVw->OnEnergyLandscape();
					return TRUE;
					
				case IDC_BTNUNDO : 
					pVw->OnBtnUndo();
					return TRUE;

				case IDC_BTN_CLASSES : 
					pVw->OnBtnClasses();
					return TRUE;

				case IDC_BTNSAVE : 
					pVw->OnSave();
					return TRUE;
				}
				
			return FALSE;

		//----- mouse buttons
		case WM_RBUTTONDOWN :		//-- add a state or rate 
			return pVw->RMouseDown( LOWORD(lParam), HIWORD(lParam) ) ? TRUE : FALSE;

		case WM_LBUTTONDOWN :		//-- Start of drag operation			
			return pVw->LMouseDown( LOWORD(lParam), HIWORD(lParam) ) ? TRUE : FALSE;

		case WM_MOUSEMOVE :			//-- Mouse move during drag operation
			if( pVw->m_pCapture == NULL ) 
				return FALSE;
			pVw->MouseMove( LOWORD(lParam), HIWORD(lParam) );
			return TRUE;

		case WM_RBUTTONUP :
		case WM_LBUTTONUP :			//-- end of drag operation
			if( pVw->m_pCapture == NULL ) 
				return FALSE;
			pVw->MouseUp( LOWORD(lParam), HIWORD(lParam) );
			return TRUE;

		case WM_DESTROY :
			pVw->OnExit();
			
			delete pVw;
			SetWindowLong( GetParent(hDlg), 0, 0 ) ;
			WindowListRemove( hDlg );
			return TRUE;

		default : 
			break;
		}
		
	return FALSE;
	} 


